/* This is a function which takes the data written under a CIP loop_
 * directive and places it in a dynamically allocated data structure.
 *
 * This function is supposed to be used as follows:  
 *
 * The calling function hits a 'loop_' directive and passes control
 * to this function.  The first lines of the CIF file below the
 * loop_ directive are should be cif data names.  Function returns
 * NULL if there was a memory allocation failure or if the cif 
 * category string is not matched.
 */
#define _POSIX_SOURCE

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* local headers */
#include "ProcessCIFLoop.h"

struct cif_loop *ProcessCIFLoop( FILE *cif, char *category )
{
   size_t len;
   char cif_line_read[CIF_LINE_LEN]; 

   char *ptr;                          /* general pointer to char for various op's */
   char (*tmp)[];                      /* pointer to an array of char's for field_names
                                        * member in the loop structure used for realloc()'s.
                                        */
#ifdef INCL_OFFSETS
   int *tmp_offset = NULL,
       *tmp_int;                       /* for moving pointers to data after realloc()'s */      
#endif       
   struct data *tmp_data_ptr;

/* make it a pointer to a structure so that the pointer value can be
 * returned to the caller.
 */
   struct cif_loop *loop;

   loop = malloc( sizeof(struct cif_loop) );
   if( !loop ) return NULL;

/* initialize the loop structure */
   loop->n_fields = 0;
   loop->n_rows = 0;
   loop->field_names = NULL;
   loop->data_values = NULL;

   while( NULL != fgets(cif_line_read, CIF_LINE_LEN, cif) ) {
      /* squash newline and squash everything after comment (#) symbol */
       if( NULL != (ptr = strrchr(cif_line_read, '\n')) ) *ptr = '\0';
       if( NULL != (ptr = strrchr(cif_line_read, '#')) ) *ptr = '\0';

       ptr = cif_line_read;

       while( *ptr ) { 
           if( isspace(*ptr) ) {  /* skip the leading white space */
               ptr++;
               continue;
           }
      
           if( UNDER_SCORE == *ptr ) {         /* Then we are in a cif data name */
               char *tmp_ptr = ptr;
               if( 0 != strncmp(ptr, category, strlen(category)) ) {
                 free( loop );
                 return NULL;
               }
               loop->n_fields++;
               while( *ptr ) {
                 if( isspace( *ptr ) ) break;
                 *ptr = tolower( *ptr );
                 ptr++;
               }
               len = ptr - tmp_ptr;     /* gives the length of the CIF data name */

               tmp = realloc( loop->field_names,
                      loop->n_fields * CIF_DATA_NAME_LEN );
               if( !tmp ) MemoryFailure( __FILE__, __LINE__ );
               loop->field_names = tmp;

               if( CIF_DATA_NAME_LEN <= len ) {
                  strncpy(loop->field_names[loop->n_fields-1], tmp_ptr, CIF_DATA_NAME_LEN - 1);
                  loop->field_names[loop->n_fields-1][CIF_DATA_NAME_LEN-1] = '\0'; 
                  fprintf( stderr, "Data name exceeds CIF specification."
                           " Truncating to %s\n", loop->field_names[loop->n_fields-1] );

               }
               else { 
                  strncpy(loop->field_names[loop->n_fields-1], tmp_ptr, len );
                  loop->field_names[loop->n_fields-1][len] = '\0'; 
               }     
               continue;
           } /* end of the cif data name processing */

/* fill the loop structure with data, assume that one row in the
 * loop is on a single line in the CIF file.  This is how most 
 * CIF files are written, but I don't think it is not required by
 * the CIF standard to be as as such, and so could be a source of
 * bugs in future.
 */
           if( isprint(*ptr) ) {
               int i, k, j = 0, 
                   in_value = 0;
               loop->n_rows++;
               i = loop->n_rows - 1;   /* to neaten up the code */
               tmp_data_ptr = realloc( loop->data_values, sizeof(struct cif_loop) +
                              loop->n_rows * ( sizeof(struct data) + CIF_LINE_LEN ) );

               if( !tmp_data_ptr ) MemoryFailure( __FILE__, __LINE__ );

               loop->data_values = tmp_data_ptr;

#ifdef  INCL_OFFSETS
               tmp_int = realloc( tmp_offset, loop->n_rows * sizeof(int) );
               if( !tmp_int ) MemoryFailure( __FILE__, __LINE__ );
               tmp_offset = tmp_int;
               if ( i > 0 ) {
                  for( k = 0; k <= i; k++ )
                       tmp_offset[k] = loop->data_values - loop->data_values[k].val_ptr;
               }
#endif
               loop->data_values[i].val_ptr = malloc( loop->n_fields * sizeof(char **) );
               if( !loop->data_values[i].val_ptr ) MemoryFailure( __FILE__, __LINE__ );

               memset( loop->data_values[i].cif_line, '\0', CIF_LINE_LEN );
               strcpy( loop->data_values[i].cif_line, ptr );
               ptr = loop->data_values[i].cif_line;
               while( *ptr ) {
                 if( !isspace(*ptr) && !in_value ) {
                     loop->data_values[i].val_ptr[j] = ptr;
                     in_value = 1;
                     j++;
                 }
                 else if( isspace(*ptr) ) {
                     in_value = 0;
                     *ptr = '\0';
                 }
                 else if( in_value ) {
                     ptr++;
                     continue;
                 }
                 ptr++;      
               }
           }
           ptr++;
       } /* end the processing of cif_line_read */

   }

#ifdef INCL_OFFSETS
   free( tmp_offset );
#endif
   return loop; 
}

/* we definately need a way to free this loop structure cleanly. So here
 * is a function to do that.   Get to this later.
 */

void FreeCIFLoop( struct cif_loop *loop ) {
    int i;
    for( i = loop->n_rows - 1; i >= 0; i-- ) {
         if( loop->data_values[i].val_ptr )
             free( loop->data_values[i].val_ptr );
    }
    if( loop->field_names ) free( loop->field_names );
    free( loop );
    return;
}

void MemoryFailure( char *file_name, int line_number ) {
     fprintf( stderr, "Memory Allocation Error: %s:%d. Exiting.\n", 
              file_name, line_number );
     exit( EXIT_FAILURE );
}
